/* This file is part of Chunky Monkey IRCD 
 *
 * Chunky Monkey IRCD is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Chunky Monkey IRCD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Chunky Monkey IRCD; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * $Id: client.c,v 1.23.4.13 2002/09/06 19:52:46 lloydy Exp $
 *
 */
#include "client.h"
#include "channel.h"
#include "ircd.h"

/* Pointer to root node */
static ClientInfo root = NULL;

/* Static routines */
static int shrink_chanlist(ClientInfo);
static int enlarge_chanlist(ClientInfo);
static int introduce(ClientInfo, ServerInfo);
static int client_netsplit_channels(ClientInfo, int, char *);
static ClientInfo client_netsplit_find(ServerInfo, ClientInfo);
static int internal_servnotice(ClientInfo, char *);

/* 
* Add a client to the tree 
* Interface to add_client_struct
*/
int client_add(const char *nick, const char *user, const char *host)
{
    ClientInfo client = NULL;

    /* We have checking setup */
    client = ircd_malloc(sizeof(struct _clientinfo));

    strncpy(client->nick, nick, NICK_LENGTH);
    strncpy(client->user, user, USER_LENGTH);
    strncpy(client->host, host, HOST_LENGTH);
    client->chan_list = calloc(10, sizeof(ChannelInfo));
    client->chanlist_no = 10;
    return client_add_by_struct(client);

}

/*
* Add a client structure into the tree
*/
int client_add_by_struct(ClientInfo client)
{
    ClientInfo current = NULL;
    int str_cmp;

    /* Set some variables */
    client->left = NULL;
    client->right = NULL;
    client->parent = NULL;

    current = root;

    /* Actually add client to tree */
    if (root == NULL) {
	/* If there are no records, add at root */
	root = client;
	client->parent = NULL;
	return 1;
    } else {
	/* Else get to where we need to be */
	while (1) {
	    str_cmp = lc_strncmp(client->nick, current->nick, NICK_LENGTH);
	    if (str_cmp > 0) {
		if (current->right == NULL) {
		    current->right = client;
		    client->parent = current;
		    return 1;
		} else {
		    current = current->right;
		}
	    }
	    if (str_cmp < 0) {
		if (current->left == NULL) {
		    current->left = client;
		    client->parent = current;
		    return 1;
		} else {
		    current = current->left;
		}
	    }
	    if (str_cmp == 0) {
		return 0;
	    }
	}
    }
}

/* 
* Delete a leaf node 
*/
static int rm_leaf_node(ClientInfo parent, ClientInfo remove)
{
    if (root != remove) {
	if (parent->right == remove) {
	    parent->right = NULL;
	} else {
	    parent->left = NULL;
	}
    } else {
	root = NULL;
    }

    return 1;
}

/* 
* delete a node with only one child 
*/
static int rm_single_child(ClientInfo parent,
			   ClientInfo remove, ClientInfo child)
{
    if (root != remove) {
	if (parent->left == remove) {
	    parent->left = child;
	} else {
	    parent->right = child;
	}

	child->parent = parent;
    } else {
	root = child;
	child->parent = NULL;
    }

    return 1;
}

/* 
* delete a node with two children	
*/
static int rm_two_kids(ClientInfo parent, ClientInfo remove)
{
    ClientInfo current = NULL;
    ClientInfo replacement = NULL;

    /* Get a replacement */
    current = remove->left;
    while (!replacement) {
	if (current->right) {
	    current = current->right;
	} else {
	    replacement = current;
	}
    }

    /* Move left subtree of replacement to
       * parents position in tree */
    if (replacement->parent->left == replacement) {
	replacement->parent->left = replacement->left;
    } else {
	replacement->parent->right = replacement->left;
    }
    if (replacement->left) {
	replacement->left->parent = replacement->parent;
    }

    /* Move right subtree of remove to the
       * replacement */
    replacement->right = remove->right;
    if (replacement->right)
	replacement->right->parent = replacement;

    /* Move the replacement to the right place */
    replacement->parent = parent;
    if (parent) {
	if (parent->left == remove) {
	    parent->left = replacement;
	} else {
	    parent->right = replacement;
	}
    } else {
	/* Root node */
	root = replacement;
    }

    /* Move left tree to right place */
    if (remove->left == replacement) {
	replacement->left = NULL;
    } else {
	replacement->left = remove->left;
	if (replacement->left)
	    replacement->left->parent = replacement;
    }


    return 1;
}

/* 
* Function to delete a node
* Calls other functions to actually delete
*/
int client_remove(const char *nick)
{

    ClientInfo current = NULL;
    int str_res;

    /* check tree actually contains something */
    if (root == NULL || nick[0] == '\0')
	return 0;

    current = root;
    while (1) {
	str_res = lc_strncmp(nick, current->nick, NICK_LENGTH);

	if (str_res == 0) {
	    /* Found the correct item */
	    if (current->left == NULL && current->right == NULL) {
		/* leaf node */
		if (rm_leaf_node(current->parent, current)) {
		    client_free(current);
		}
		return 1;
	    }

	    if (current->left != NULL && current->right == NULL) {
		/* Node with a left child */
		if (rm_single_child
		    (current->parent, current, current->left)) {
		    client_free(current);
		}
		return 1;
	    }

	    if (current->right != NULL && current->left == NULL) {
		/* Node with a right child */
		if (rm_single_child
		    (current->parent, current, current->right)) {
		    client_free(current);
		}
		return 1;
	    }

	    if (current->right != NULL && current->left != NULL) {
		/* node with two children */
		if (rm_two_kids(current->parent, current)) {
		    client_free(current);
		}
		return 1;
	    }
	}
	/* Move pointers along */
	if (str_res < 0)
	    current = current->left;

	if (str_res > 0)
	    current = current->right;

	/* Drop out if we reach the end */
	if (current == NULL)
	    return 0;

    }
}

/* 
 * Function to find a specified user 
 * 
 */
ClientInfo client_locate(const char *user)
{
    ClientInfo current = NULL;
    int str_cmp;

    current = root;

    if (current == NULL || user == NULL)
	return NULL;

    while (1) {

	str_cmp = lc_strncmp(user, current->nick, NICK_LENGTH);

	if (str_cmp == 0)
	    return current;

	if (str_cmp > 0) {
	    if (current->right == NULL)
		return NULL;
	    current = current->right;
	}
	if (str_cmp < 0) {
	    if (current->left == NULL)
		return NULL;
	    current = current->left;
	}
    }
}

/* 
* Unlink a client from the tree, don't free() 
*/
int client_unlink(const char *nick)
{

    ClientInfo current = NULL;
    int str_res;

    /* check tree actually contains something */
    if (root == NULL || nick[0] == '\0')
	return 0;

    current = root;
    while (1) {
	str_res = lc_strncmp(nick, current->nick, NICK_LENGTH);

	if (str_res == 0) {
	    /* Found the correct item */
	    if (current->left == NULL && current->right == NULL)
		/* leaf node */
		return rm_leaf_node(current->parent, current);


	    if (current->left != NULL && current->right == NULL)
		/* Node with a left child */
		return rm_single_child(current->parent,
				       current, current->left);

	    if (current->right != NULL && current->left == NULL)
		/* Node with a right child */
		return rm_single_child(current->parent,
				       current, current->right);


	    if (current->right != NULL && current->left != NULL)
		/* node with two children */
		return rm_two_kids(current->parent, current);
	}
	/* Move pointers along */
	if (str_res < 0)
	    current = current->left;

	if (str_res > 0)
	    current = current->right;

	/* Drop out if we reach the end */
	if (current == NULL)
	    return 0;

    }
}

/*
* Free a client structure 
*/
int client_free(ClientInfo client)
{
    if (client->chan_list != NULL) {
	free(client->chan_list);
	client->chan_list = NULL;
    }
    if (client->away_msg) {
	free(client->away_msg);
	client->away_msg = NULL;
    }

    free(client);

    return 1;
}

/* ******* Client Keeps a Channel Listing ******** */
/*        These routines maintain this list        */

/* 
* Enlarges channel list size, needs a little work 
*/
static int enlarge_chanlist(ClientInfo client)
{
    ChannelInfo *temp_ptr = NULL;
    int i;
    temp_ptr =
	realloc(client->chan_list,
		(client->chanlist_no + 10) * sizeof(ChannelInfo));

    if (temp_ptr == NULL) {
	return 0;
    } else {
	/* Realloc frees the original block */
	client->chan_list = temp_ptr;
	client->chanlist_no += 10;
	/* Realloc doesn't initialise memory */
	for (i = client->chanlist_no - 10; i < client->chanlist_no; i++) {
	    client->chan_list[i] = NULL;
	}
	return 1;
    }
}

/*
* Add a channel pointer to a client structure
*/
int client_add_chan(ClientInfo client, ChannelInfo chan)
{
    int i;

    if (client->chan_list == NULL) {
	client->chan_list = calloc(10, sizeof(ChannelInfo));
	if (client->chan_list == NULL) {
	    return 1;
	}
	client->chanlist_no = 10;
    }

    /* Find first empty or end */
    for (i = 0;
	 i < client->chanlist_no && client->chan_list[i] != NULL; i++);

    if (i == client->chanlist_no) {
	if (enlarge_chanlist(client) == 0) {
	    return 1;
	}
	i = client->chanlist_no - 10;
    }
    client->chan_list[i] = chan;
    client->chan_count++;
    return 1;

}

/* 
* Remove channel from a client list 
*/
int client_rem_chan(ClientInfo client, ChannelInfo chan)
{
    int i, j;

    for (i = 0; i < client->chanlist_no; i++) {
	if (client->chan_list[i] == chan) {
	    /* Search from end to find replacement */
	    for (j = client->chanlist_no - 1; j > i; j--) {
		if (client->chan_list[j] != NULL) {
		    client->chan_list[i] = client->chan_list[j];
		    client->chan_list[j] = NULL;
		    shrink_chanlist(client);
		    client->chan_count--;
		    return 1;
		}
	    }
	    if (j == i) {
		client->chan_list[i] = NULL;
		shrink_chanlist(client);
		client->chan_count--;
		return 1;
	    }
	}
    }

    /* do we need to shrink the list? */
    shrink_chanlist(client);

    return 1;
}

/* 
* shrink client channel list if needed 
*/
static int shrink_chanlist(ClientInfo client)
{
    int i, non_null = 0;
    ChannelInfo *temp_ptr = NULL;

    for (i = client->chanlist_no - 10;
	 i < client->chanlist_no && non_null == 0; i++) {
	if (client->chan_list[i] != NULL) {
	    non_null = 1;
	}
    }

    if (non_null == 0 && client->chanlist_no > 10) {
	temp_ptr =
	    realloc(client->chan_list,
		    (client->chanlist_no - 10) * sizeof(ChannelInfo));
	if (temp_ptr == NULL) {
	    return 0;
	} else {
	    /* No need to free() original block, realloc does it */
	    client->chan_list = temp_ptr;
	    client->chanlist_no -= 10;
	}
    }
    return 1;

}



/* 
* send our client list to a newly-connected server
*/
int clients_introduce(ServerInfo s_ptr)
{
    /* Call the recursive bit */
    introduce(root, s_ptr);

    return 1;
}

/* 
* recursive algorithm for sending client information to 
* a newly-connected server 
*/
static int introduce(ClientInfo node, ServerInfo s_ptr)
{
    char buffer[BUF_SIZE];

    if (!node)
	return 0;

    /* Recursive Bit */
    if (node->left) {
	introduce(node->left, s_ptr);
    }
    if (node->right) {
	introduce(node->right, s_ptr);
    }

    /* No Children, send this node */
    ircd_sprintf(buffer, BUF_SIZE, "NICK %s %d %ld %s %s %s %s 0 :%s",
		 node->nick, node->server->hops, node->ts,
		 get_mode_str(node->mode), node->user, node->host,
		 node->server->name, node->real_name);
    send_origin_server(s_ptr, NULL, buffer);


    return 1;
}

/* 
* Find a client connected to the dud server
* send a message, then remove them from the tree
*/
int clients_netsplit(ServerInfo src, ServerInfo split,
		     char *orig_server_name)
{
    char buffer[BUF_SIZE];
    ClientInfo c_ptr = NULL;
    int rand_no;

    /* Find a client connected via that server */
    while ((c_ptr = client_netsplit_find(split, root))) {
	rand_no = rand() % 1000 + 1;
	ircd_sprintf(buffer, BUF_SIZE, "QUIT :%s %s",
		     get_my_server()->name, orig_server_name);

	/* Deal with channels */
	client_netsplit_channels(c_ptr, rand_no, buffer);
	/* Report it to servers (only non-ts4 servers */
	server_report_netsplit(src, c_ptr->nick, buffer);

	if (c_ptr->mode & MODE_IRCOP)
	    server_state.ops--;
	client_remove(c_ptr->nick);

    }

    return 1;
}

/*
 * Process the client's channels
 * For each channel, report it to users, and clean up
 */
static int client_netsplit_channels(ClientInfo client,
				    int rand_no, char *quit_msg)
{
    int i = 0;
    ChannelMember curr_user = NULL;

    for (i = 0;
	 i < client->chanlist_no && client->chan_list[i] != NULL; i++) {
	/* remove from channel user list */
	channel_rem_client(client, client->chan_list[i]);

	/* Send it */
	curr_user = client->chan_list[i]->users_start;
	while (curr_user) {
	    if (curr_user->user_info->fd != -1
		&& curr_user->user_info->rand_no != rand_no) {
		send_client_client(curr_user->user_info, client, quit_msg);
		curr_user->user_info->rand_no = rand_no;
	    }
	    curr_user = curr_user->next;
	}

	/* Close channel if needed */
	channel_needclose(client->chan_list[i]);

	/* Set pointer to NULL */
	client->chan_list[i] = NULL;

    }
    return 0;
}

/* 
* Find a client connected to a particular server
* Return a pointer to the client, or NULL
*/
static ClientInfo client_netsplit_find(ServerInfo split, ClientInfo node)
{
    ClientInfo current = NULL;

    if (!node)
	return NULL;

    if (node->server == split) {
	current = node;
    }

    if (node->left != NULL && current == NULL)
	current = client_netsplit_find(split, node->left);

    if (node->right != NULL && current == NULL)
	current = client_netsplit_find(split, node->right);

    return current;

}

/* 
* Implements KILL using QUIT,
* Might not be correct, so I'll change it later
*/
int client_kill(ServerInfo s_ptr,	/* Server the KILL request came from */
		ServerInfo s_src,	/* Server source .. or */
		ClientInfo c_src,	/* Client source */
		ClientInfo r_ptr,	/* Client to kill */
		char *reason)
{				/* Reason */
    char buffer[BUF_SIZE];
    int i;
    int path_len = 0;
    char *origin = NULL;
    char path[301] = "\0";
    ServerInfo curr_ptr = NULL;
    ServerInfo my_ptr = NULL;

    my_ptr = get_my_server();

    /* Set the origin */
    if (s_src) {
	origin = s_src->name;
    } else {
	origin = c_src->nick;
	curr_ptr = c_src->server;
    }

    /* Get the path */
    ircd_sprintf(path, 300, "%s", origin);
    while (curr_ptr != NULL) {
	path_len = strlen(path);
	strncat(path, "!", 300 - path_len);
	path_len++;
	strncat(path, curr_ptr->name, 300 - path_len);

	curr_ptr = curr_ptr->via_ptr;
    }

    /* Tell client */
    /* Only done if the dead client is local */
    if (r_ptr->fd != -1) {
	ircd_sprintf(buffer, BUF_SIZE, "KILL %s :%s", r_ptr->nick, path);
	send_origin_client(r_ptr, origin, buffer);
	ircd_sprintf(buffer, BUF_SIZE, "ERROR :Closing Link: %s by %s, (Killed (%s))", r_ptr->nick, origin, reason ? reason : "");	/* Handle NULL reason */
	send_origin_client(r_ptr, NULL, buffer);

	/* Remove the client from the fdlist */
	ircd_fdlist_clear_entry(r_ptr->fd);

	close(r_ptr->fd);
    }

    /* Inform the channel */
    ircd_sprintf(buffer, BUF_SIZE, "QUIT :Killed by %s - %s", origin,
		 reason ? reason : "");

    for (i = 0; i < r_ptr->chanlist_no && r_ptr->chan_list[i] != NULL; i++) {
	/* remove from channel user list */
	channel_rem_client(r_ptr, r_ptr->chan_list[i]);

	/* Inform channel */
	send_to_local_channel(r_ptr, r_ptr->chan_list[i], buffer);

	/* Close channel if needed */
	channel_needclose(r_ptr->chan_list[i]);

	/* Set pointer to NULL */
	r_ptr->chan_list[i] = NULL;
    }

    ircd_sprintf(buffer, BUF_SIZE, "KILL %s :%s", r_ptr->nick, reason);
    servers_relay(s_ptr, origin, buffer);

    /* Server Notice */
    ircd_sprintf(buffer, BUF_SIZE,
		 "Recieved Kill for %s.  From %s Path:%s (%s)",
		 r_ptr->nick, origin, path, reason);
    clients_servnotice(buffer);

    /* Reduce Counter and Remove from tree */
    r_ptr->server->client_count--;
    client_remove(r_ptr->nick);

    return 1;
}

/*
* Generate and send a names list for
* the clients not on a channel
*/
int clients_nochan_names(ClientInfo c_ptr, ClientInfo curr_node,
			 char *passed_buffer)
{
    char buffer[BUF_SIZE];

    /* Called from outside this file, start at root */
    if (!curr_node) {
	strcpy(buffer, "* * :");
	clients_nochan_names(c_ptr, root, buffer);

	send_numeric(c_ptr, get_my_server(), 353, buffer);
    } else {
	if ((curr_node->chan_list == NULL
	     || curr_node->chan_list[0] == NULL)
	    && !(curr_node->mode & MODE_INVIS)) {

	    if (strlen(curr_node->nick) <
		(BUF_SIZE - 2 - strlen(passed_buffer))) {
		strcat(passed_buffer, " ");
		strcat(passed_buffer, curr_node->nick);
	    } else {
		send_numeric(c_ptr, get_my_server(), 353, passed_buffer);
		strcpy(passed_buffer, "* * :");
		strcat(passed_buffer, curr_node->nick);
	    }

	}
	if (curr_node->left)
	    clients_nochan_names(c_ptr, curr_node->left, passed_buffer);
	if (curr_node->right)
	    clients_nochan_names(c_ptr, curr_node->right, passed_buffer);
    }

    return 1;
}

/*
* Send a message to all clients with a particular mode set
*/
int clients_servnotice(char *message)
{
    return internal_servnotice(root, message);
}

/* 
* Recursive routine to do the dirty work 
*/
static int internal_servnotice(ClientInfo node, char *message)
{
    char buffer[BUF_SIZE] = { "\0" };

    /* If we get called with node==NULL, use root */
    if (node) {

	if (node->left)
	    internal_servnotice(node->left, message);

	if (node->right)
	    internal_servnotice(node->right, message);

	if ((node->mode & MODE_SERVNOT) && (node->fd != -1)) {
	    ircd_sprintf(buffer, BUF_SIZE,
			 "NOTICE %s :*** Notice -- %s",
			 node->nick, message);
	    send_origin_client(node, get_my_server()->name, buffer);
	}
    }

    return 1;
}

/* 
 * Used to get a client that matches a mask 
 * in nick,host,user,realname or server name
 */
int client_who_mask(char *mask,
		    ClientInfo src_ptr, ClientInfo node, int start)
{
    char buffer[BUF_SIZE];
    int match = 0;
    ChannelInfo chan_ptr = NULL;
    ChannelMember cli_info = NULL;
    char mode_char = ' ';

    if (start == 1)
	node = root;

    if (!node) {
	return 0;
    }

    /* Try and match */
    if (check_string(node->host, mask))
	match = 1;

    if ((!match) && check_string(node->server->name, mask))
	match = 1;

    if ((!match) && check_string(node->user, mask))
	match = 1;

    if ((!match) && check_string(node->real_name, mask))
	match = 1;

    if ((!match) && check_string(node->nick, mask))
	match = 1;

    /* if we match, send to query client */
    if (match) {
	if (node->chan_list && node->chan_list[0]) {
	    chan_ptr = node->chan_list[0];
	    cli_info = channel_is_client_on(node, chan_ptr);
	    if (cli_info->chan_mode & CHANUSR_OP) {
		mode_char = '@';
	    } else {
		if (cli_info->chan_mode & CHANUSR_VOICE) {
		    mode_char = '+';
		} else {
		    mode_char = ' ';
		}
	    }

	} else {
	    chan_ptr = NULL;
	}

	ircd_sprintf(buffer, BUF_SIZE,
		     "%s %s %s %s %s %c%c%s%d %s",
		     chan_ptr ? chan_ptr->name : "*", node->user,
		     node->host, node->server->name, node->nick,
		     node->mode & MODE_AWAY ? 'G' : 'H', mode_char,
		     mode_char == ' ' ? ":" : " :",
		     node->server->hops, node->real_name);
	send_numeric(src_ptr, get_my_server(), 352, buffer);

    }

    /* Try child nodes */
    client_who_mask(mask, src_ptr, node->left, 0);
    client_who_mask(mask, src_ptr, node->right, 0);

    return 1;
}

/*
 * Send a message to all the clients that are on a specific server
 */
void client_msg_by_server_mask(ClientInfo origin,	/* Origin of this message (op or server) */
			       char *mask,	/* The mask for the clients we want */
			       char *type,	/* Either PRIVMSG or NOTICE */
			       char *msg,	/* The message we're sending */
			       ClientInfo node)
{				/* The node we're dealing with */
    ClientInfo curr_node = NULL;
    char buffer[BUF_SIZE];

    /* Use the right node */
    if (node) {
	curr_node = node;
    } else {
	curr_node = root;
    }

    /* Check this mask */
    /* We only send to local clients */
    if (check_string(curr_node->server->name, mask)
	&& (curr_node->fd != -1)) {

	ircd_sprintf(buffer, BUF_SIZE, "%s %s :%s", type,
		     curr_node->nick, msg);

	if (origin) {
	    send_client_client(curr_node, origin, buffer);
	}
    }


    /* Recurse */
    if (curr_node->left) {
	client_msg_by_server_mask(origin, mask, type, msg,
				  curr_node->left);
    }

    if (curr_node->right) {
	client_msg_by_server_mask(origin, mask, type, msg,
				  curr_node->right);
    }

}

/*
 * Send a message to all the clients that match a specifc host mask
 */
void client_msg_by_host_mask(ClientInfo origin,	/* Origin of this message (op or server) */
			     char *mask,	/* The mask for the clients we want */
			     char *type,	/* Either PRIVMSG or NOTICE */
			     char *msg,	/* The message we're sending */
			     ClientInfo node)
{				/* The node we're dealing with */
    ClientInfo curr_node = NULL;
    char buffer[BUF_SIZE];

    /* Use the right node */
    if (node) {
	curr_node = node;
    } else {
	curr_node = root;
    }

    /* Check this mask */
    /* Only send to local clients */
    if (check_string(curr_node->host, mask) && (curr_node->fd != -1)) {

	ircd_sprintf(buffer, BUF_SIZE, "%s %s :%s", type,
		     curr_node->nick, msg);

	if (origin) {
	    send_client_client(curr_node, origin, buffer);
	}
    }


    /* Recurse */
    if (curr_node->left) {
	client_msg_by_server_mask(origin, mask, type, msg,
				  curr_node->left);
    }

    if (curr_node->right) {
	client_msg_by_server_mask(origin, mask, type, msg,
				  curr_node->right);
    }

}
