/* This file is part of Chunky Monkey IRCD 
 *
 * Chunky Monkey IRCD is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Chunky Monkey IRCD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Chunky Monkey IRCD; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * $Id: network.c,v 1.31.4.44 2002/11/03 22:29:37 lloydy Exp $
 *
 */
#include "network.h"

/* Utility Routines */
static int send_server(ServerInfo, char *, size_t length);
static int send_raw(int, char *, size_t length);
#ifdef USE_OPENSSL
static int send_ssl_raw(SSL *, char *, size_t length);
static int send_encrypt_raw(int, EncryptInfo, char *, size_t length);
#endif
static ServerInfo reduce_to_local(ServerInfo);

/* Recving routines */
#ifdef USE_OPENSSL
static int ircd_encrypt_recv(int, char *, EncryptInfo);
static int ircd_ssl_recv(int, char *, SSL *);
#endif
static int ircd_plaintext_recv(int, char *);

/* 
* called to setup tcp socket, returns socket or -1
*/
int setup_tcp(struct sockaddr_in *server)
{
    int tcp_socket, sock_val;

    /* Get TCP Socket */
    if ((tcp_socket = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
	perror("socket");
	return -1;
    }

    /* Set reuse */
    sock_val = 1;
    setsockopt(tcp_socket, SOL_SOCKET, SO_REUSEADDR, &sock_val,
	       sizeof(int));

    /* Bind to socket */
    if (bind
	(tcp_socket, (struct sockaddr *) server,
	 sizeof(struct sockaddr)) == -1) {
	close(tcp_socket);
	perror("bind");
	return -1;
    }

    /* Listen on socket */
    if (listen(tcp_socket, NET_MAXLISTEN) == -1) {
	close(tcp_socket);
	perror("listen");
	return -1;
    }

    /* Set non-blocking */
    fcntl(tcp_socket, F_SETFL, O_NONBLOCK);

    log_msg(LOG_DEBUG1, "Successfully bound to %s:%d",
	    inet_ntoa(server->sin_addr), ntohs(server->sin_port));

    return tcp_socket;

}

/* 
* send a message to a client, add the necessary stuff depending
* on whether they are local or not 
*/
int send_client_client(ClientInfo dest_ptr, ClientInfo src_ptr,
		       char *message)
{
    char buffer[512];
    /* message is NICK timmy, or suchlike */

    if (dest_ptr->fd != -1) {
	/* Local recipient */
	if (src_ptr) {
	    if (ircd_sprintf
		(buffer, 512, ":%s!%s@%s %s\r\n",
		 src_ptr->nick, src_ptr->user, src_ptr->host, message)) {
		return -1;
	    }
	} else {
	    if (ircd_sprintf(buffer, 512, "%s\r\n", message)) {
		return -1;
	    }
	}
#ifdef USE_OPENSSL
	if (dest_ptr->ssl) {
	    return send_ssl_raw(dest_ptr->ssl, buffer, strlen(buffer));
	} else {
	    return send_raw(dest_ptr->fd, buffer, strlen(buffer));
	}
#else
	return send_raw(dest_ptr->fd, buffer, strlen(buffer));
#endif
    } else {
	if (ircd_sprintf
	    (buffer, 512, ":%s %s\r\n", src_ptr->nick, message)) {
	    return -1;
	} else {
	    return send_server(dest_ptr->server, buffer, strlen(buffer));
	}
    }
}

/*
*  Send a client messages to a server
*/
int send_origin_server(ServerInfo dest, char *origin, char *message)
{
    char buffer[512];

    if (ircd_sprintf
	(buffer, 512, "%s%s%s%s\r\n", origin ? ":" : "",
	 origin ? origin : "", origin ? " " : "", message))
	return -1;

    return send_server(dest, (void *) buffer, strlen(buffer));
}

/*
* Send something to a temp
*/
int send_temp(struct tempinfo *dest, char *message)
{
    char buffer[512];

    if (ircd_sprintf
	((char *) buffer, 512, ":%s %s\r\n", get_my_server()->name,
	 message))
	return -1;

    log_msg(LOG_DEBUG2, "Sending %s on temp", buffer);

#ifdef USE_OPENSSL
    if (dest->enc_ctx) {
	return send_encrypt_raw(dest->fd, dest->enc_ctx, buffer,
				strlen(buffer));
    } else if (dest->ssl) {
	return SSL_write(dest->ssl, buffer, strlen((char *) buffer));
    } else {
	return send_raw(dest->fd, buffer, strlen((char *) buffer));
    }
#else
    return send_raw(dest->fd, buffer, strlen((char *) buffer));
#endif

}

/* 
* Send origin only to client
*/
int send_origin_client(ClientInfo dest, char *origin, char *message)
{
    char buffer[512];

    if (ircd_sprintf
	(buffer, 512, "%s%s%s%s\r\n", origin ? ":" : "",
	 origin ? origin : "", origin ? " " : "", message))
	return -1;

    if (dest->fd == -1) {
	return send_server(dest->server, buffer, strlen(buffer));
    } else {
#ifdef USE_OPENSSL
	if (dest->ssl) {
	    return send_ssl_raw(dest->ssl, buffer, strlen(buffer));
	} else {
	    return send_raw(dest->fd, buffer, strlen(buffer));
	}
#else
	return send_raw(dest->fd, buffer, strlen(buffer));
#endif
    }

}

/* 
* Sends raw output to a socket 
*/
static int send_raw(int fd, char *buffer, size_t length)
{
/*	log_msg(LOG_INFO,"Sent: %s",buffer); */
    return send(fd, buffer, length, 0);
}

/*
 * Send raw output to a SSL socket
 */
#ifdef USE_OPENSSL
static int send_ssl_raw(SSL * ctx, char *buffer, size_t length)
{
    return SSL_write(ctx, buffer, length);
}

/*
 * Encrypt output to an fd
 */
static int send_encrypt_raw(int fd, EncryptInfo ctx, char *buffer,
			    size_t length)
{
    unsigned char enc_buffer[512];
    int enc_length;
    char len_buf[5];

    /* Avoid warning */
    length = length;

    /* Encrypt */
    encrypt_buffer(ctx, (unsigned char *) buffer, enc_buffer, &enc_length);

    /* Send the length */
    ircd_sprintf(len_buf, 5, "%04d", enc_length);
    len_buf[4] = '\0';
    send(fd, len_buf, 4, 0);

    /* Send the encrypted data */
    return send(fd, enc_buffer, enc_length, 0);

}
#endif


/* 
* Send to a server, recursive 
*/
static int send_server(ServerInfo s_ptr, char *buffer, size_t length)
{
    if (s_ptr->fd == -1) {
	return send_server(s_ptr->via_ptr, buffer, length);
    } else {
#ifdef USE_OPENSSL
	if (s_ptr->enc_ctx) {
	    return send_encrypt_raw(s_ptr->fd, s_ptr->enc_ctx, buffer,
				    length);
	} else {
	    return send_raw(s_ptr->fd, buffer, length);
	}
#else
	return send_raw(s_ptr->fd, buffer, length);
#endif
    }

}

/* 
* sends a numeric reply 
* only local clients can get this sent to them 
*/
int send_numeric(ClientInfo dest, ServerInfo src, int code, char *text)
{
    char buffer[BUF_SIZE];

    if (!dest) {
	return -1;
    }

    ircd_sprintf(buffer, 480, "%03d %s %s", code, dest->nick, text);

    return send_origin_client(dest, src->name, buffer);
}

/* 
* Loop though a channel and send a message 
* Sends to clients on other servers
*/
int send_to_channel(ClientInfo sender, ChannelInfo chan, char *message)
{
    ChannelMember curr_client = NULL;

    curr_client = chan->users_start;

    while (curr_client) {
	if (curr_client->user_info != sender) {
	    /* Clients on another server are dealt with by that server */
	    send_client_client(curr_client->user_info, sender, message);
	}
	curr_client = curr_client->next;
    }

    return 1;
}

/*
 * Only send to servers with members of this channel on
 */
int send_to_channel_selective(ClientInfo sender, ChannelInfo chan,
			      ServerInfo s_ptr, char *buffer)
{
    int rand_no;
    ChannelMember curr_client = NULL;

    rand_no = rand() % 1000 + 1;

    curr_client = chan->users_start;
    while (curr_client) {
	if (curr_client->user_info != sender) {
	    if (curr_client->user_info->fd == -1) {
		ServerInfo local_route = NULL;

		local_route =
		    reduce_to_local(curr_client->user_info->server);

		/* Remote client */
		if (local_route->rand_no != rand_no) {
		    /* We've not already sent it */
		    if (local_route != s_ptr) {
			send_origin_server
			    (curr_client->
			     user_info->server, sender->nick, buffer);
			local_route->rand_no = rand_no;
		    }
		}

	    } else {
		send_client_client(curr_client->user_info, sender, buffer);
	    }
	}
	curr_client = curr_client->next;
    }

    return 1;

}

/*
 * Reduce a [very] remote server into a local one 
 * for use in routing decisions
 */
static ServerInfo reduce_to_local(ServerInfo remote)
{
    ServerInfo my_serv = NULL;
    ServerInfo curr_serv = NULL;

    /* Get my Server */
    my_serv = get_my_server();

    curr_serv = remote;

    while (curr_serv->via_ptr != my_serv) {
	curr_serv = curr_serv->via_ptr;
    }

    return curr_serv;
}


/* 
* Only send to local channel members
*/
int send_to_local_channel(ClientInfo sender, ChannelInfo chan,
			  char *message)
{
    ChannelMember curr_client = NULL;

    curr_client = chan->users_start;

    while (curr_client) {
	if (curr_client->user_info != sender
	    && curr_client->user_info->fd != -1) {
	    /* Clients on another server are dealt with by that server */
	    send_client_client(curr_client->user_info, sender, message);
	}
	curr_client = curr_client->next;
    }

    return 1;
}

/*
* Send to local channel ops
*/
int send_to_local_channel_ops(ClientInfo sender, ChannelInfo chan,
			      char *message)
{
    ChannelMember curr_client = NULL;

    curr_client = chan->users_start;

    while (curr_client) {
	if (curr_client->user_info != sender
	    && curr_client->user_info->fd != -1
	    && (curr_client->chan_mode & CHANUSR_OP)) {
	    send_client_client(curr_client->user_info, sender, message);
	}
	curr_client = curr_client->next;
    }

    return 1;
}

/* Non-basic routines - all depend on routines above */
/* Basically, they could be in cmd_*.c, but due to the */
/* dynamic stuff, and the fact other routines rely on them */
/* they are in here */

/* Send MOTD to a client */
int send_motd(ClientInfo c_ptr)
{
    char buffer[BUF_SIZE];
    int i;
    ServerInfo m_ptr = NULL;

    m_ptr = get_my_server();

    if (server_config.motd == NULL) {
	/* Send a numeric to tell them */
	send_numeric(c_ptr, m_ptr, 422, ":MOTD File is missing");
    } else {
	/* Send the motd */
	ircd_sprintf(buffer, BUF_SIZE,
		     ":- %s Message of the day -", m_ptr->name);
	send_numeric(c_ptr, m_ptr, 375, buffer);

	for (i = 0; server_config.motd[i] != NULL; i++) {
	    ircd_sprintf(buffer, BUF_SIZE, ":- %s", server_config.motd[i]);
	    send_numeric(c_ptr, m_ptr, 372, buffer);
	}

	send_numeric(c_ptr, m_ptr, 376, ":End of MOTD Command");
    }

    return 1;
}

/* 
* Send a ping
*/
int send_ping(struct fd_wrap *fd_wrapper)
{
    char buffer[BUF_SIZE];

    if (fd_wrapper->server) {
	ircd_sprintf(buffer, BUF_SIZE, "PING %s %s\r\n",
		     get_my_server()->name, fd_wrapper->server->name);
    } else {
	ircd_sprintf(buffer, BUF_SIZE, "PING %s\r\n",
		     get_my_server()->name);
    }

    if (fd_wrapper->server) {
#ifdef USE_OPENSSL
	if (fd_wrapper->server->enc_ctx) {
	    return send_encrypt_raw(fd_wrapper->server->fd,
				    fd_wrapper->server->enc_ctx,
				    buffer, strlen(buffer));
	} else {
	    return send_raw(fd_wrapper->server->fd, buffer,
			    strlen(buffer));
	}
#else
	return send_raw(fd_wrapper->server->fd, buffer, strlen(buffer));
#endif
    }
    if (fd_wrapper->client) {
#ifdef USE_OPENSSL
	if (fd_wrapper->client->ssl) {
	    return send_ssl_raw(fd_wrapper->client->ssl,
				buffer, strlen(buffer));
	} else {
	    return send_raw(fd_wrapper->client->fd,
			    buffer, strlen(buffer));
	}
#else
	return send_raw(fd_wrapper->client->fd, buffer, strlen(buffer));
#endif
    }
    if (fd_wrapper->temp) {
#ifdef USE_OPENSSL
	if (fd_wrapper->temp->ssl) {
	    return send_ssl_raw(fd_wrapper->temp->ssl,
				buffer, strlen(buffer));
	} else if (fd_wrapper->temp->enc_ctx) {
	    return send_encrypt_raw(fd_wrapper->temp->fd,
				    fd_wrapper->temp->enc_ctx,
				    buffer, strlen(buffer));
	} else {
	    return send_raw(fd_wrapper->temp->fd, buffer, strlen(buffer));
	}
#else
	return send_raw(fd_wrapper->temp->fd, buffer, strlen(buffer));
#endif
    }

    return -1;
}

/*
 * Try and receive from a local fd
 */
int ircd_recv(struct fd_wrap *fd, char *buffer)
{

    if (fd->server) {
#ifdef USE_OPENSSL
	if (fd->server->enc_ctx) {
	    return ircd_encrypt_recv(fd->server->fd, buffer,
				     fd->server->enc_ctx);
	} else {
	    return ircd_plaintext_recv(fd->server->fd, buffer);
	}
#else
	return ircd_plaintext_recv(fd->server->fd, buffer);
#endif
    }
    if (fd->temp) {
#ifdef USE_OPENSSL
	if (fd->temp->enc_ctx) {
	    return ircd_encrypt_recv(fd->temp->fd, buffer,
				     fd->temp->enc_ctx);
	} else if (fd->temp->ssl) {
	    return ircd_ssl_recv(fd->temp->fd, buffer, fd->temp->ssl);
	} else {
	    return ircd_plaintext_recv(fd->temp->fd, buffer);
	}
#else
	return ircd_plaintext_recv(fd->temp->fd, buffer);
#endif
    }
    if (fd->client) {
#ifdef USE_OPENSSL
	if (fd->client->ssl) {
	    return ircd_ssl_recv(fd->client->fd, buffer, fd->client->ssl);
	} else {
	    return ircd_plaintext_recv(fd->client->fd, buffer);
	}
#else
	return ircd_plaintext_recv(fd->client->fd, buffer);
#endif
    }
    return -1;
}

/*
 * Handle server recv's
 */
#ifdef USE_OPENSSL
static int ircd_encrypt_recv(int fd, char *buffer, EncryptInfo ctx)
{
    /* Encrypted */
    char enc_len[5];
    int len = 0, clean_len = 0;
    unsigned char enc_buffer[1024];
    int retval = 0;


    while (len == 0 && (retval = recv(fd, enc_len, 4, 0)) == 4) {
	enc_len[4] = '\0';
	len = 0;

	if (enc_len[0] > 47 && enc_len[0] < 58 &&
	    enc_len[1] > 47 && enc_len[0] < 58 &&
	    enc_len[2] > 47 && enc_len[0] < 58 &&
	    enc_len[3] > 47 && enc_len[0] < 58) {
	    len = atoi(enc_len);
	    log_msg(LOG_DEBUG2, "Got %d bytes, from %s", len, enc_len);
	}
    }

    if (len) {
	/* If we don't get all the bytes we want,
	 * we have to loop tightly here */

	while (recv(fd, enc_buffer, len, MSG_PEEK) != len) {
	    usleep(10);		/* Might be a little out */
	}
	recv(fd, enc_buffer, len, 0);
	clean_len =
	    decrypt_buffer(ctx, enc_buffer, (unsigned char *) buffer, len);
	buffer[clean_len] = '\0';

    }

    if (clean_len > 0) {
	return clean_len;
    } else {
	if (retval == 0) {
	    errno = ECONNRESET;
	}
	return -1;
    }
}
#endif

/*
 * Handle PLaintext
 */
static int ircd_plaintext_recv(int fd, char *buffer)
{
    /* Plain text */
    int found_end = 0, i = 0;
    int numbytes = 0;

    numbytes = recv(fd, buffer, 2048, MSG_PEEK);
    for (i = 0; i < numbytes && !found_end; i++) {
	if (buffer[i] == '\n' || buffer[i] == '\r') {
	    if (buffer[i + 1] == '\n' || buffer[i + 1] == '\r') {
		i++;
	    }
	    found_end = 1;
	}
    }

    if (found_end) {
	numbytes = recv(fd, buffer, i, 0);
	buffer[numbytes] = '\0';
	return numbytes;
    } else {
	if (numbytes == 0) {
	    log_msg(LOG_DEBUG2,
		    "Got a connection reset in ircd_plaintext_recv");
	    errno = ECONNRESET;
	    numbytes = -1;
	}
	return numbytes;
    }
}

#ifdef USE_OPENSSL
/*
 * Handle ssl recv
 */
static int ircd_ssl_recv(int fd, char *buffer, SSL * ssl)
{
    int numbytes = 0, i = 0, found_end = 0;

    fd = fd;

    numbytes = SSL_peek(ssl, buffer, 2048);
    for (i = 0; i < numbytes && !found_end; i++) {
	if (buffer[i] == '\n' || buffer[i] == '\r') {
	    if (buffer[i + 1] == '\n' || buffer[i + 1] == '\r') {
		i++;
	    }
	    found_end = 1;
	}
    }

    if (found_end) {
	numbytes = SSL_read(ssl, buffer, i);
	buffer[numbytes] = '\0';
	return numbytes;
    } else {
	if (numbytes == 0) {
	    log_msg(LOG_DEBUG2, "Got a connection reset in ircd_ssl_recv");
	    errno = ECONNRESET;
	    numbytes = -1;
	}
	return numbytes;
    }

}
#endif

/* 
* Send a channel name list to a client 
*/
int send_namelists(ChannelInfo chan, ClientInfo client)
{
    char start_char = '=';
    char buffer[BUF_SIZE];
    char next[BUF_SIZE];
    char nicks[BUF_SIZE];
    char start[BUF_SIZE];
    ChannelMember curr_client = NULL;
    ChannelMember on_chan = NULL;
    unsigned int avail_chars = 0;

    /* Check to see if person who called is on channel */
    on_chan = channel_is_client_on(client, chan);

    /* decide on start char */
    switch (chan->mode & (CMODE_SECRET | CMODE_PRIVATE)) {

    case (CMODE_SECRET):
	start_char = '@';
	break;

    case (CMODE_PRIVATE):
	start_char = '*';
	break;

    default:
	start_char = '=';
	break;
    }

    /* Start to formulate list */
    ircd_sprintf(start, BUF_SIZE, "%c %s :", start_char, chan->name);
    avail_chars = BUF_SIZE - 1 - strlen(start);
    nicks[0] = '\0';

    /* Sweep through channel */
    if (!(chan->mode & CMODE_SECRET) || (on_chan)) {
	curr_client = chan->users_start;

	while (curr_client != NULL) {

	    if (!(curr_client->user_info->mode & MODE_INVIS)
		|| (on_chan)) {

		ircd_sprintf(next, BUF_SIZE, "%s%s ",
			     (curr_client->
			      chan_mode & CHANUSR_OP) ? "@"
			     : (curr_client->chan_mode & CHANUSR_HALFOP)
			     ? "%" : (curr_client->
				      chan_mode &
				      CHANUSR_VOICE) ? "+"
			     : "", curr_client->user_info->nick);
		if (strlen(next) < avail_chars) {
		    strcat(nicks, next);
		    avail_chars -= strlen(next);
		} else {
		    ircd_sprintf(buffer, BUF_SIZE, "%s%s", start, nicks);
		    send_numeric(client, get_my_server(), 353, buffer);
		    avail_chars =
			BUF_SIZE - 1 - strlen(start) - strlen(next);
		    strcpy(nicks, next);
		}

	    }

	    curr_client = curr_client->next;
	}

	/* Send list */
	ircd_sprintf(buffer, BUF_SIZE, "%s%s", start, nicks);
	send_numeric(client, get_my_server(), 353, buffer);
    }

    return 1;
}

/*
* Send a LUSERS set of replies when prompted
* by a LUSERS command or at client registration
*/
int send_lusers(ClientInfo c_ptr, char *mask)
{
    char buffer[BUF_SIZE];
    ServerInfo my_server = NULL;
    long int global_clients;
    long int global_servers;

    /* Get stuff we need */
    my_server = get_my_server();
    servers_global_counts(mask, &global_clients, &global_servers);

    /* Send LUSERCLIENT */
    ircd_sprintf(buffer, BUF_SIZE,
		 ":There are %ld users and %0d services on %ld servers",
		 global_clients, 0, global_servers);
    send_numeric(c_ptr, my_server, 251, buffer);

    /* Send LUSEROP */
    if (server_state.ops) {
	ircd_sprintf(buffer, BUF_SIZE, "%0ld :operators online",
		     server_state.ops);
	send_numeric(c_ptr, my_server, 252, buffer);
    }

    /* Send LUSERUNKNOWN */
    if (server_state.temp_count) {
	ircd_sprintf(buffer, BUF_SIZE, "%ld :unknown clients",
		     server_state.temp_count);
	send_numeric(c_ptr, my_server, 253, buffer);
    }

    /* Send LUSERCHANNELS */
    if (server_state.channels) {
	ircd_sprintf(buffer, BUF_SIZE, "%ld :channels formed",
		     server_state.channels);
	send_numeric(c_ptr, my_server, 254, buffer);
    }

    /* Send LUSERME */
    ircd_sprintf(buffer, BUF_SIZE,
		 ":I Have %ld clients and %ld servers",
		 my_server->client_count, my_server->server_count);
    send_numeric(c_ptr, my_server, 255, buffer);

    return 1;
}

/* 
* Force a peer to disconnect 
*/
int force_quit(int fd, char *reason)
{
    char buffer[BUF_SIZE];
    int i;
    struct fd_wrap *wrapper = NULL;

    wrapper = ircd_fdlist_get_entry(fd);

    if (!wrapper) {
	return -1;
    }


    if (wrapper->temp) {
	ircd_sprintf(buffer, BUF_SIZE, "ERROR :Closing Link :%s", reason);
	log_msg(LOG_DEBUG1, "Closed Temporary link - %s", reason);
	send_temp(wrapper->temp, buffer);

#ifdef USE_OPENSSL
	if (wrapper->temp->ssl) {
	    SSL_shutdown(wrapper->temp->ssl);
	    SSL_free(wrapper->temp->ssl);
	    wrapper->temp->ssl = NULL;
	}
#endif
	/* Just dequeue and close port */
	ircd_fdlist_clear_entry(fd);


	temp_rem(fd);
	close(fd);
	return 0;
    }

    if (wrapper->client) {

	/* Force kill client */
	ircd_sprintf(buffer, BUF_SIZE, "QUIT :%s", reason);

	if (wrapper->client->chan_list != NULL) {
	    for (i = 0;
		 i < wrapper->client->chanlist_no
		 && wrapper->client->chan_list[i] != NULL; i++) {
		/* remove from channel user list */
		channel_rem_client(wrapper->client,
				   wrapper->client->chan_list[i]);

		/* Inform channel */
		send_to_channel(wrapper->client,
				wrapper->client->chan_list[i], buffer);

		/* Close channel if needed */
		channel_needclose(wrapper->client->chan_list[i]);

		/* Set pointer to NULL */
		wrapper->client->chan_list[i] = NULL;
	    }
	}

	/* Announce to Servers */
	servers_announce(wrapper->client->nick, buffer);

	/* Report we're closing the link */
	ircd_sprintf(buffer, BUF_SIZE, "ERROR :Closing Link :%s", reason);
	send_origin_client(wrapper->client, NULL, buffer);

	if ((wrapper->client->mode & MODE_IRCOP))
	    server_state.ops--;
	get_my_server()->client_count--;

#ifdef USE_OPENSSL
	if (wrapper->client->ssl) {
	    SSL_shutdown(wrapper->client->ssl);
	    SSL_free(wrapper->client->ssl);
	    wrapper->client->ssl = NULL;
	}
#endif
	/* Remove from tree */
	client_remove(wrapper->client->nick);

	/* Clear up fdlist */
	ircd_fdlist_clear_entry(fd);
	close(fd);

	return 1;
    }

    if (wrapper->server) {
	/* Log it */
	log_msg(LOG_BASIC,
		"Broke Local Server link to %s. Reason: %s",
		wrapper->server->name, reason);

	/* Inform other servers */
	ircd_sprintf(buffer, BUF_SIZE,
		     "Received a SQUIT for %s, reason %s",
		     wrapper->server->name, reason);
	clients_servnotice(buffer);

	/* Work out clients on that server */
	clients_netsplit(NULL, wrapper->server, wrapper->server->name);

	/* Work out which clients and servers behind this one where disconnected */
	server_netsplit(wrapper->server, wrapper->server, reason,
			wrapper->server->name);

	/* Tell other servers we netsplitted from this server, and they work out the rest */
	ircd_sprintf(buffer, BUF_SIZE, "SQUIT %s :%s",
		     wrapper->server->name, reason);
	servers_relay(wrapper->server, get_my_server()->name, buffer);

	server_hash_rem(wrapper->server->name);
	get_my_server()->server_count--;

	server_free(wrapper->server);

	ircd_fdlist_clear_entry(fd);
	close(fd);

	return 1;


    }

    return 0;

}
