/* This file is part of Chunky Monkey IRCD 
 *
 * Chunky Monkey IRCD is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Chunky Monkey IRCD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Chunky Monkey IRCD; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * $Id: temp.c,v 1.2.4.5 2002/09/06 19:52:46 lloydy Exp $
 *
 */
#include "temp.h"

/* temp.c - Functions to maintain temp list used to store clients between 
 * when they connect and when they identify themselves as a server or a client */

/* Start of list */
static struct tempinfo *list_start = NULL;

static void temp_free(TempInfo);

/*
 * Add a peer to the temporary list
 * All peers, including clients and servers are put in the list
 * until they register
 */
struct tempinfo *temp_add(char *host, int fd)
{
    struct tempinfo *new = NULL;
    struct tempinfo *curr = NULL;
    struct tempinfo *prev = NULL;

    /* Attempt to allocate space for it */
    new = ircd_malloc(sizeof(struct tempinfo));

    /* Fill with details */
    new->fd = fd;
    strncpy(new->host, host, HOST_LENGTH);
    new->nick[0] = '\0';
    new->real_name[0] = '\0';
    new->flags = 0;
    new->user[0] = '\0';
    new->name[0] = '\0';
    new->info[0] = '\0';
    new->next = NULL;

    /* Add */
    prev = NULL;
    curr = list_start;

    while (curr) {
	prev = curr;
	curr = curr->next;
    }

    if (!prev) {
	list_start = new;
    } else {
	prev->next = new;
    }

    server_state.temp_count++;
    return new;
}

/*
 * Remove a client from the list
 * can take either the hostname or the fd to find the client
 * using the fd is more efficent. 
 */
int temp_rem(int fd)
{
    struct tempinfo *curr = NULL, *temp = NULL, *prev = NULL;

    if (!list_start)
	return 0;

    if (list_start->fd == fd) {
	temp = list_start;
	list_start = list_start->next;

	temp_free(temp);

	server_state.temp_count--;
	return 1;
    } else {

	prev = list_start;
	curr = list_start->next;

	while (curr && curr->fd != fd) {
	    prev = curr;
	    curr = curr->next;
	}

	if (curr) {
	    temp = curr;
	    prev->next = curr->next;
	    temp_free(temp);
	    server_state.temp_count--;
	    return 1;
	}
    }
    return 0;
}

/* 
 * Locate a client in the temporary list using the nick
 * only used to check for duplicate nicks
 */
int temp_locate(char *nick)
{
    struct tempinfo *curr = NULL;

    if (list_start == NULL) {
	return 0;
    }

    curr = list_start;
    while (curr != NULL) {
	if (lc_strncmp(curr->nick, nick, -1) == 0) {
	    return 1;
	} else {
	    curr = curr->next;
	}
    }


    return 0;
}

/*
 * Free a temp structure
 */
void temp_free(TempInfo freeme)
{
#ifdef USE_OPENSSL
    if (freeme->enc_ctx) {
	free(freeme->enc_ctx);
    }
/*	if(freeme->ssl) {
		SSL_free(freeme->ssl);
	} */
#endif
    free(freeme);
}
