/* This file is part of Chunky Monkey IRCD 
 
 * Chunky Monkey IRCD is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Chunky Monkey IRCD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Chunky Monkey IRCD; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * $Id: utility.c,v 1.22.2.1.2.9 2002/09/06 19:52:46 lloydy Exp $
 *
 */
#include "utility.h"

static const char *get_next_substr(const char *);

/* 
* allocates a string that can be free'd when finished with 
*/
char *string_alloc(int length)
{

    return calloc(length + 1, sizeof(char));

}

/* 
* interface to malloc with some extra checks
*/
void *ircd_malloc(size_t size)
{
    void *retval = NULL;

    /* Try to get the memory */
    retval = malloc(size);

    if (!retval) {
	log_msg(LOG_BASIC,
		"Failed to allocate memory, reason from malloc() %s",
		strerror(errno));
	exit(1);
    } else {
	memset(retval, 0, size);
	return retval;
    }

}


/* 
* replacement for strcmp 
* lowercase's it before comparing 
* treats { as lc [,} as lc ],| as lc \ and ^ as lc ~ 
*/
int lc_strncmp(const char *string1, const char *string2, int limit)
{
    char *buffer1;
    char *buffer2;
    int i;
    int retval;

    /* Check strings are valid */
    if (string1 == NULL || string2 == NULL)
	return EXIT_FAILURE;


    buffer1 = string_alloc(strlen(string1));
    buffer2 = string_alloc(strlen(string2));

    if (buffer1 == NULL || buffer2 == NULL) {
	free(buffer1);
	free(buffer2);
	return EXIT_FAILURE;
    }
    /* Lowercase strings */
    for (i = 0; string1[i] != '\0'; i++) {
	if (string1[i] > 64 && string1[i] < 95) {
	    buffer1[i] = string1[i] + 32;
	} else {
	    buffer1[i] = string1[i];
	}
    }
    buffer1[i] = '\0';

    for (i = 0; string2[i] != '\0'; i++) {
	if (string2[i] > 64 && string2[i] < 95) {
	    buffer2[i] = string2[i] + 32;
	} else {
	    buffer2[i] = string2[i];
	}
    }
    buffer2[i] = '\0';

    /* Call strcmp */
    if (limit != -1) {
	retval = strcmp(buffer1, buffer2);
    } else {
	retval = strncmp(buffer1, buffer2, limit);
    }

    free(buffer1);
    free(buffer2);

    return retval;
}

/* 
 * Routine to handle differences between different
 * vsnprintf() implementations, returns -1 if overflow
 * 0 otherwise
*/
int ircd_sprintf(char *buffer, int buf_size, char *format, ...)
{
    va_list items;
    int retval = 0;

    va_start(items, format);
    retval = vsnprintf(buffer, buf_size, format, items);

    if (retval == -1 || retval >= buf_size) {
	/* Overflow Occured */
	return -1;
    } else {
	/* Success */
	return 0;
    }
}

/* 
* Check format of a nick 
*/
int check_nick(char *nick)
{
    int i;


    /* Check first chars */
    switch (nick[0]) {
    case 48:
    case 49:
    case 50:
    case 51:
    case 52:
    case 53:
    case 54:
    case 55:
    case 56:
    case 57:
    case 45:
	/* These aren't allowed */
	return 0;
	break;
    default:
	break;
    }


    /* Check individual characters */
    for (i = 0; nick[i] != '\0'; i++) {
	/* Huge if structure */
	if ((nick[i] > 47 && nick[i] < 58) ||
	    (nick[i] > 64 && nick[i] < 125) || (nick[i] == 45)) {

	} else {
	    /* Invalid char */
	    return 0;
	}
    }

    /* Return true if nick is ok */
    return 1;
}

/* 
* Check strings with * in 
* Used by bans and server 
* Return 1 on match, 0 if not
* Strings Must be \0 terminated
*/

int check_string(const char *string, const char *mask)
{
    int i, j;
    char *next_part = NULL;

    for (i = 0, j = 0; string[i] != '\0'; i++) {

	if (string[i] != mask[j]) {
	    if (mask[j] == '*') {

		j++;

		if (mask[j] == '\0') {
		    return 1;
		} else {
		    /* Try to skip to a good place */
		    next_part =
			strstr(&(string[i]), get_next_substr(&(mask[j])));

		    if (next_part) {
			return check_string(next_part, &(mask[j]));
		    } else {
			return 0;
		    }
		}

	    } else {
		return 0;
	    }
	} else {
	    j++;
	}
    }

    return 1;
}

/*
* Called to get the next token to find
*/
static const char *get_next_substr(const char *string)
{

    const char *start = NULL;
    const char *end = NULL;
    static char *buffer = NULL;
    int i;

    free(buffer);
    buffer = NULL;

    start = string;

    for (i = 0; string[i] != '\0'; i++) {
	if (string[i] == '*') {
	    end = &(string[i - 1]);
	}
    }

    if (end) {
	buffer = calloc(end - start + 2, sizeof(char));
	strncpy(buffer, start, end - start + 1);
	return buffer;
    } else {
	return start;
    }


}


/* 
* when given a mode numeric, is able to work
* out which mode string to us
*/
char *get_mode_str(int mode)
{
    static char retval[10];
    int i = 0;

    retval[i++] = '+';

    if (mode & MODE_IRCOP)
	retval[i++] = 'o';

    if (mode & MODE_SERVNOT)
	retval[i++] = 's';

    if (mode & MODE_INVIS)
	retval[i++] = 'i';

    if (mode & MODE_WALLOPS)
	retval[i++] = 'w';

    if (mode & MODE_REGIST)
	retval[i++] = 'r';

    if (mode & MODE_TRUSTED)
	retval[i++] = 't';

    retval[i++] = '\0';

    return retval;
}


/*
* Fill a string with the current channel mode 
*/
char *get_chan_mode_str(ChannelInfo chan, char addrem)
{
    char modeline[15];
    char params[50] = "\0";
    char limit[10];
    static char buffer[60];

    int i = 0;

    if (chan->mode == 0) {
	return NULL;
    }

    modeline[i++] = addrem;

    if (chan->mode & CMODE_TOPIC)
	modeline[i++] = 't';

    if (chan->mode & CMODE_MEMONLYSEND)
	modeline[i++] = 'n';

    if (chan->mode & CMODE_SECRET)
	modeline[i++] = 's';

    if (chan->mode & CMODE_MODERATED)
	modeline[i++] = 'm';

    if (chan->mode & CMODE_SERVREG)
	modeline[i++] = 'r';

    if (chan->mode & CMODE_REGONLY)
	modeline[i++] = 'R';

    if (chan->mode & CMODE_TRUSTONLY)
	modeline[i++] = 'T';

    if (chan->mode & CMODE_KEY) {
	modeline[i++] = 'k';
	strncpy(params, chan->key, 50);
    }

    if (chan->user_limit != -1) {
	modeline[i++] = 'l';
	if (addrem == '+') {
	    ircd_sprintf(limit, 10, "%d", chan->user_limit);
	    strcat(params, limit);
	}
    }

    modeline[i] = '\0';

    ircd_sprintf(buffer, 60, "%s %s", modeline, params);

    return buffer;

}
