/* This file is part of Chunky Monkey IRCD 
 *
 * Chunky Monkey IRCD is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Chunky Monkey IRCD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Chunky Monkey IRCD; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * $Id: cmd_connect.c,v 1.19.4.9 2002/09/06 19:52:46 lloydy Exp $
 *
 */
#include <errno.h>
#include "server.h"
#include "chunkyircd.h"
#include "resolver.h"
#include "ircd.h"
#include "modules.h"

CMIRCD_MODULE(1.02, CMD_CLIENT);

static int establish_connection(char *);

/* 
 * Establish a server to server link
 * OPER's only
 */
int cmd_connect(ClientInfo c_ptr, ServerInfo s_ptr, Cmd cmd)
{
    char buffer[BUF_SIZE];

    /* Check its only from a local client */
    if (!c_ptr || s_ptr) {
	log_msg(LOG_DEBUG1, "Received a CONNECT from a non-local client");

	return 0;
    }

    /* Check the client has ops */
    if (!(c_ptr->mode & MODE_IRCOP)) {
	ircd_sprintf(buffer, BUF_SIZE,
		     ":Permission denied - You're not an IRC Operator");
	send_numeric(c_ptr, get_my_server(), 481, buffer);

	return 0;
    }

    /* Check we have enough params */
    if (!check_params(cmd, 1)) {
	/* Not enough params */
	ircd_sprintf(buffer, BUF_SIZE, "CONNECT :Not enough parameters");
	send_numeric(c_ptr, get_my_server(), 461, buffer);

	return 0;
    }

    /* If they sent a hostmask, check its for us */
    if (check_params(cmd, 2)) {
	if (!check_string(get_my_server()->name, cmd->params[1])) {
	    /* Not for us */

	    return 0;
	} else {
	    /* Its for us */
	}

    }

    /* Call a routine to establish the connection */
    ircd_sprintf(buffer, BUF_SIZE,
		 "CONNECT used by %s, attempting to link to server %s",
		 c_ptr->nick, cmd->params[0]);
    clients_servnotice(buffer);
    establish_connection(cmd->params[0]);


    return 1;
}

/* 
* establish connection
*/
static int establish_connection(char *server_name)
{
    struct sockaddr_in sin_server;
    struct in_addr temp_addr;
    struct fd_wrap new_wrapper;
    int fd = -1;
    struct tempinfo *t_ptr = NULL;
    struct cline *cline = NULL;
    char *host = NULL;
    int port;


    cline = get_cline(server_name);
    if (!cline) {
	log_msg(LOG_BASIC, "Unable to CONNECT to %s - no CLINE",
		server_name);
	return 0;
    } else {
	host = cline->host;
	port = cline->port;
    }

    memset(&(sin_server), 0, sizeof(struct sockaddr));
    memset(&(new_wrapper), 0, sizeof(struct fd_wrap));
    new_wrapper.res_id = -1;

    fd = socket(PF_INET, SOCK_STREAM, 0);

    if (fd == -1) {
	printf("Failed to get socket for connect()\n");
	perror("socket");
	exit(0);
    }

    t_ptr = temp_add(host, fd);

    /* Set non-blocking */
    fcntl(fd, F_SETFL, O_NONBLOCK);

    t_ptr->port = port;
    strncpy(t_ptr->name, server_name, SERVNAME_MAX);
    if (t_ptr->port < 1025) {
	free(t_ptr);
	return 0;
    }

    log_msg(LOG_DEBUG1,
	    "Connecting to %s using hostname %s and port %d",
	    server_name, host, port);

    if (inet_aton(host, &temp_addr)) {
	if (server_init_connect(t_ptr, host)) {
	    temp_rem(fd);
	} else {
	    new_wrapper.temp = t_ptr;
	}

	/* Submit to get a reverse */
	new_wrapper.res_id =
	    resolver_submit_reverse(t_ptr->connect_ip.s_addr);
	log_msg(LOG_DEBUG1,
		"Processing CONNECT message, issued connect() using IP %s",
		inet_ntoa(t_ptr->connect_ip));
    } else {
	/* Submit to the resolver */
	new_wrapper.temp = t_ptr;
	new_wrapper.res_id = resolver_submit(host);
	new_wrapper.temp->flags |= USR_SHOULDCONN;
	strncpy(new_wrapper.temp->connect_host, host, HOST_LENGTH);
	log_msg(LOG_DEBUG1,
		"Processing CONNECT message, submitted hostname %s for resolution before we connect",
		host);
    }

    ircd_fdlist_set_entry(fd, &new_wrapper);

    log_msg(LOG_DEBUG2, "cmd_connect() done, returning");
    return 1;
}
