/* This file is part of Chunky Monkey IRCD 
 *
 * Chunky Monkey IRCD is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Chunky Monkey IRCD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Chunky Monkey IRCD; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * $Id: cmd_notice.c,v 1.12.4.8 2002/09/06 19:52:47 lloydy Exp $
 *
 */
#include "modules.h"

/* Export version number */
CMIRCD_MODULE(1.02, CMD_CLIENT | CMD_SERVER);

static void notice_to_user(ClientInfo, ClientInfo, char *);
static void notice_to_channel(ServerInfo source, ClientInfo sender,
			      ChannelInfo chan, int ops, char *msg);

/*
* cmd_notice, deals with NOTICE routines
*/
int cmd_notice(ClientInfo c_ptr, ServerInfo s_ptr, Cmd cmd)
{
    ClientInfo recip = NULL;
    ClientInfo sender = NULL;
    ChannelInfo chan = NULL;
    char buffer[BUF_SIZE];
    int send_to_ops = 0;

    /* Check the params */
    if (!check_params(cmd, 1) || !cmd->text) {
	if (c_ptr) {
	    /* return error */
	    ircd_sprintf(buffer, BUF_SIZE,
			 "%s : Not enough parameters\n", cmd->command);
	    send_numeric(c_ptr, get_my_server(), 461, buffer);

	}
	return 0;
    }

    /* Get a sender */
    if (c_ptr) {
	sender = c_ptr;
    } else {
	sender = client_locate(cmd->origin);
    }
    if (!sender) {
	return 0;
    }

    /* Try to send to a room */
    if (cmd->params[0][0] == '@') {
	send_to_ops = 1;
	chan = locate_channel(&(cmd->params[0][1]));
    } else {
	chan = locate_channel(cmd->params[0]);
    }
    if (chan) {
	notice_to_channel(s_ptr, sender, chan, send_to_ops, cmd->text);
	return 1;
    }

    /* Try to send to user */
    recip = client_locate(cmd->params[0]);
    if (recip) {
	notice_to_user(sender, recip, cmd->text);
	return 1;
    }

    /* Try and use $* mask */
    if ((cmd->params[0][0] == '$') && (sender->mode & MODE_IRCOP)) {
	size_t max_len = 0;
	char *next_ptr = NULL;
	char trim_buffer[BUF_SIZE];

	max_len = BUF_SIZE - 14 - strlen(cmd->params[0]);
	next_ptr = cmd->text;

	while (next_ptr) {
	    if (strlen(next_ptr) > max_len) {
		strncpy(trim_buffer, next_ptr, max_len);
		next_ptr = next_ptr + max_len;
	    } else {
		strncpy(trim_buffer, next_ptr, max_len);
		next_ptr = NULL;
	    }
	    /* Send to servers */
	    ircd_sprintf(buffer, BUF_SIZE, "NOTICE %s :%s",
			 cmd->params[0], cmd->text);
	    servers_relay(s_ptr, sender->nick, buffer);
	    /* Send to local clients */
	    /* XXX - Is this really needed with new code at 1.0.1 ? */
	    client_msg_by_server_mask(sender,
				      &(cmd->params[0][1]),
				      "NOTICE", cmd->text, NULL);
	}
	return 1;
    }

    /* Try and use #* mask */
    if ((cmd->params[0][0] == '#') && (sender->mode & MODE_IRCOP)) {
	size_t max_len = 0;
	char *next_ptr = NULL;
	char trim_buffer[BUF_SIZE];

	max_len = BUF_SIZE - 14 - strlen(cmd->params[0]);
	next_ptr = cmd->text;

	while (next_ptr) {
	    if (strlen(next_ptr) > max_len) {
		strncpy(trim_buffer, next_ptr, max_len);
		next_ptr = next_ptr + max_len;
	    } else {
		strncpy(trim_buffer, next_ptr, max_len);
		next_ptr = NULL;
	    }
	    /* Send to servers */
	    ircd_sprintf(buffer, BUF_SIZE, "NOTICE %s :%s",
			 cmd->params[0], trim_buffer);
	    servers_relay(s_ptr, sender->nick, buffer);
	    /* Send to local clients */
	    /* XXX - Is this really needed with new code at 1.0.1 ? */
	    client_msg_by_host_mask(sender,
				    &(cmd->params[0][1]),
				    "NOTICE", trim_buffer, NULL);
	}
	return 1;
    }

    /* Return error to a client */
    if (c_ptr) {
	ircd_sprintf(buffer, BUF_SIZE, "%s :No such nick/channel",
		     cmd->params[0]);
	send_numeric(sender, get_my_server(), 401, buffer);
    }
    return 1;

}

/*
 * Send a notice to a user
 */
static void notice_to_user(ClientInfo sender, ClientInfo recip, char *text)
{
    char cmd_buff[BUF_SIZE];

    /* Form a string */
    if (text) {
	ircd_sprintf(cmd_buff, BUF_SIZE, "NOTICE %s :%s",
		     recip->nick, text);
	send_client_client(recip, sender, cmd_buff);
    }
}

/*
 * Send a notice to a channel
 */
static void notice_to_channel(ServerInfo relay, ClientInfo sender,
			      ChannelInfo chan, int ops_only, char *txt)
{
    char buffer[BUF_SIZE];

    /* Check they can actually send it */
    if (channel_msg_check_perms(chan, sender)) {
	if (ops_only) {
	    ircd_sprintf(buffer, BUF_SIZE, "NOTICE @%s :%s",
			 chan->name, txt);
	    send_to_local_channel_ops(sender, chan, buffer);
	    servers_relay(relay, sender->nick, buffer);
	} else {
	    size_t max_len = 0;
	    char *next_ptr = NULL;
	    char trim_buffer[BUF_SIZE];

	    max_len = BUF_SIZE - 14 - strlen(chan->name);
	    next_ptr = txt;

	    while (next_ptr) {
		if (strlen(next_ptr) > max_len) {
		    strncpy(trim_buffer, next_ptr, max_len);
		    next_ptr = next_ptr + max_len;
		} else {
		    strncpy(trim_buffer, next_ptr, max_len);
		    next_ptr = NULL;
		}
		ircd_sprintf(buffer, BUF_SIZE,
			     "NOTICE %s :%s", chan->name, trim_buffer);
		send_to_channel_selective(sender, chan, relay, buffer);
	    }
	}
    } else {
	ircd_sprintf(buffer, BUF_SIZE,
		     "%s :Cannot send to channel", chan->name);
	send_numeric(sender, get_my_server(), 404, buffer);
    }
}
